/* This file is part of the LWRPC library */
#include <lwrpc_internal.h>
#include <lwrpc.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#define lwrpc_desc_size sizeof(lwrpc_t)
extern MPI_Comm comm_out;
volatile int termination_flag = 0;

typedef union _result {
	int i;
	double d;
} result_t;

static lwrpc_t no_work_desc;

int process_a_received_descriptor(lwrpc_t *work/*, int tag1*/)
{
	int reuse = 0;
	lwrpc_t *stolen_work;
	MPI_Status status;
	int i, istat, tag;
	char *mem;

	work->next = NULL;

	tag = work->sourcevpid;
	if ((tag < 0) || (tag > MAX_CPUS)) {	
		Error1("Invalid message tag: %d", tag);
	}

	switch (work->type) {
	case LWRPC_ANSWER:
		{
		for (i = 0; i < work->narg; i++) {
			if ((work->callway[i] == CALL_BY_RES) || (work->callway[i] == CALL_BY_REF)) {
					enter_comm_cs();
					istat = MPI_Recv((void *)work->localarg[i], work->quantity[i], work->dtype[i], work->sourcenode, tag, comm_out, &status);
					leave_comm_cs();
			}
		}
		if ((work->parent)&&(_lwrpc_depsatisfy(work->parent))) {
				lwrpc_to_i_pq(work->parent);
		}
		reuse = 1;
		return reuse;
		}
		break;


	case LWRPC_NORMAL_ENQUEUE:
		{
		if (work->homenode == lwrpc_node_id()) {
			lwrpc_to_i_rq(work);
			reuse = 0;
			return reuse;
		}
		else {
			receive_arguments(work, tag);

			if (work->insert_private) {
					if (work->insert_in_front)
						lwrpc_to_i_pq(work);
					else
						lwrpc_to_i_pq_end(work);
			}
			else {
				if (work->target_queue == -1) {
					if (work->insert_in_front)
						lwrpc_to_i_rq(work);
					else
						lwrpc_to_i_rq_end(work);
				}
				else {
					if (work->insert_in_front)
						lwrpc_to_i_lrq(work->target_queue, work);
					else
						lwrpc_to_i_lrq_end(work->target_queue, work);
				}
			}

			reuse = 0;
			return reuse;
		}
	}
	break;

	case TERMINATE_LOCAL_SERVER_THREAD:
		pthread_exit(0);
		reuse = 1;
		return reuse;
		break;

		
	case TERMINATE_WORKER_THREADS:
		termination_flag = 1;
		thread_signal();
		if (work->localarg[0] != lwrpc_node_id()) {
			appl_finished++;
		}

		reuse = 1;
		return reuse;
		break;

	case SPMD_BARRIER:
		barrier_arrived++;
		reuse = 1;
		return reuse;
		break;

	case RESET_STATISTICS:
		lwrpc_reset_statistics();
		break;

	default:
/*		Error1("Unkown descriptor type on node %d", lwrpc_node_id());*/
		break;
	}

	reuse = 1;
	return reuse;
}


pthread_mutex_t stc_lock = PTHREAD_MUTEX_INITIALIZER;
static int server_thread_created;

void *server_loop (void *arg)
{
	lwrpc_t *work;
	int reuse = 0;
	MPI_Status status;
	int istat;

	memset(&no_work_desc, 0, sizeof(lwrpc_t));
	no_work_desc.type = LWRPC_NO_WORK;

	pthread_mutex_lock(&stc_lock);
	server_thread_created = 1;
	pthread_mutex_unlock(&stc_lock);

	while (1) {			
		if (!reuse)
			work = _lwrpc_get_reused_desc();

		memset(work, 0, sizeof(lwrpc_t));

		if (!non_blocking) {
			istat = MPI_Recv(work, lwrpc_desc_size, MPI_CHAR, MPI_ANY_SOURCE, MAX_CPUS, comm_out, &status);
		}
		else {
			int flag = 0;
			MPI_Request request;

			enter_comm_cs();
			istat = MPI_Irecv(work, lwrpc_desc_size, MPI_CHAR, MPI_ANY_SOURCE, MAX_CPUS, comm_out, &request);
			leave_comm_cs();
			while (1) {
				enter_comm_cs();
				MPI_Test(&request, &flag, &status);
				leave_comm_cs();
				if (appl_finished == tobe_finished) pthread_exit(0);
				if (flag == 1) {
					break;
				}
				else {
					thread_sleep(0);
				}
			}
		}

		reuse = process_a_received_descriptor(work /*,0 */);
		if (reuse) {
			_lwrpc_put_reused_desc(work);
			reuse = 0;
		}

	}

	return 0;
}


static pthread_mutex_t server_thread_lock = PTHREAD_MUTEX_INITIALIZER;
static int server_thread_alive = 0;

void start_server_thread()
{
	int res;
	pthread_attr_t attr;

	pthread_attr_init(&attr);
	pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM);
	pthread_mutex_lock(&server_thread_lock);

	if (server_thread_alive == 1) {
		pthread_mutex_unlock(&server_thread_lock);
		return; 
	}

/*	pthread_mutex_lock(&stc_lock);*/
	server_thread_created = 0;
/*	pthread_mutex_unlock(&stc_lock);*/

	res = pthread_create(&server_thread, &attr, server_loop, NULL);
	if (res != 0) {
		Error("server thread was not created!\n");
	}
	while (1) {
/*		pthread_mutex_lock(&stc_lock);*/
		if (server_thread_created==1) {
/*			pthread_mutex_unlock(&stc_lock);*/
			break;
		}
		else {
/*			pthread_mutex_unlock(&stc_lock);*/
			sched_yield();
		}
	}

	server_thread_alive = 1;

	pthread_mutex_unlock(&server_thread_lock);

}

void shutdown_server_thread()
{
	lwrpc_t mydata;
	
	pthread_mutex_lock(&server_thread_lock);
	if (server_thread_alive == 0) {
		pthread_mutex_unlock(&server_thread_lock);
		return; 
	}

	memset(&mydata, 0, sizeof(mydata));
	send_descriptor(lwrpc_node_id(), &mydata, TERMINATE_LOCAL_SERVER_THREAD);
	pthread_join(server_thread, NULL);
	server_thread_alive = 0;
	
	pthread_mutex_unlock(&server_thread_lock);

	
}

void terminate_workers()
{
	lwrpc_t mydata;
	int node, nnodes = lwrpc_num_nodes();
	
	memset(&mydata, 0, sizeof(lwrpc_t));
	mydata.localarg[0] = lwrpc_node_id();
	mydata.homenode = lwrpc_node_id();

	for (node = 0; node < nnodes; node++) {
		if (node != lwrpc_node_id())
			send_descriptor(node, &mydata, TERMINATE_WORKER_THREADS);
	}
}

void notify_spmd_barrier()
{
	lwrpc_t mydata;
	int node, nnodes = lwrpc_num_nodes();
	
	memset(&mydata, 0, sizeof(lwrpc_t));
	mydata.localarg[0] = lwrpc_node_id();
	mydata.homenode = lwrpc_node_id();

	for (node = 0; node < nnodes; node++) {
		if (node != lwrpc_node_id())
			send_descriptor(node, &mydata, SPMD_BARRIER);
	}
}


void memsetvalue(void *mem, double value, int n, MPI_Datatype dtype)
{
	int i;
	int *ig;
	double *dg;

	if (value == 0) {
		int typesize;
		MPI_Type_size(dtype, &typesize);
		memset(mem, 0, n*typesize);
		return;
	}

	if ((dtype == MPI_INT)||(dtype == MPI_INTEGER)) {
		ig = (int *)mem;
		for (i = 0; i < n; i++) ig[i] = (int)value;
	}
	else if ((dtype == MPI_DOUBLE)||(dtype == MPI_DOUBLE_PRECISION)) {
		dg = (double *)mem;
		for (i = 0; i < n; i++) dg[i] = value;
	}
	else {
		printf("UNSUPPORTED DATATYPE\n"); fflush(0);
	}
}

void lwrpc_reset_statistics ()
{
	lwrpc_t mydata;
	int node, nnodes = lwrpc_num_nodes();
    
	memset(&mydata, 0, sizeof(lwrpc_t));
	for (node = 0; node < nnodes; node++) {
		if (node != lwrpc_node_id())
			send_descriptor(node, &mydata, RESET_STATISTICS);
		else
			_lwrpc_reset_statistics();
	}
}

void F77_FUNC_ (lwrpc_reset_statistics, LWRPC_RESET_STATISTICS) ()
{
	lwrpc_reset_statistics();
}
